-- Copyright (C) 2014 Bradley Smith <brad@brad-smith.co.uk>
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.

local NexusMeter = {
  combats = { }
}
local ApolloUnit = Unit
local L

local pairs = pairs
local ipairs = ipairs
local tinsert = table.insert
local tremove = table.remove
local tsort = table.sort
local max = math.max
local min = math.min
local round = function(val) return math.floor(val + .5) end
local setmetatable = setmetatable

local Event = Event

local Events = {}

NexusMeter.Units = {}
local Units = NexusMeter.Units
NexusMeter.Abilities = {}
local Abilities = NexusMeter.Abilities

local StopTracking = false -- used by EndCombatAfterKill
local LastDamageAction = 0
local LastUpdate = 0
local LastTimerUpdate = 0
local InCombat = false
local NeedsUpdate = false
local Permanent = false -- manual combat start
local UnitAvailabilityQueue = {}
NexusMeter.CurrentCombat = {}

local FilteredAbilities = {
--	[L["Ability name"]] = true,
}
local EndCombatAfterKill = {
--  "id",
}
local NewCombatAfterKill = {
--	"id",
}

NexusMeter.settings = {
	windows = {},
	classColors = {},
	abilityTypeColors = {},
	lock = false,
	alwaysShowPlayer = false,
	showOnlyBoss = false,
	showScrollbar = false,
	showRankNumber = true,
	showPercent = true,
	showAbsolute = true,
	showShortNumber = false,
	mergeAbilitiesByName = true, -- Review this, do we actually need it or is there a bug?
	opacity = 0,
	tooltips = true,
	enabled = true,
}

-- Class colors taken directly from website
NexusMeter.settings.classColors[GameLib.CodeEnumClass.Warrior] = {0.8, 0.1, 0.1}
NexusMeter.settings.classColors[GameLib.CodeEnumClass.Engineer] = {0.65, 0.65, 0}
NexusMeter.settings.classColors[GameLib.CodeEnumClass.Esper] = {0.1, 0.5, 0.7}
NexusMeter.settings.classColors[GameLib.CodeEnumClass.Medic] = {0.2, 0.6, 0.1}
NexusMeter.settings.classColors[GameLib.CodeEnumClass.Stalker] = {0.5, 0.1, 0.8}
NexusMeter.settings.classColors[GameLib.CodeEnumClass.Spellslinger] = {0.9, 0.4, 0}

NexusMeter.settings.abilityTypeColors[GameLib.CodeEnumDamageType.Magic] = {0.5, 0.1, 0.8}
NexusMeter.settings.abilityTypeColors[GameLib.CodeEnumDamageType.Tech] = {0.2, 0.6, 0.1}
NexusMeter.settings.abilityTypeColors[GameLib.CodeEnumDamageType.Physical] = {0.6, 0.6, 0.6}
NexusMeter.settings.abilityTypeColors[GameLib.CodeEnumDamageType.HealShields] = {0.1, 0.5, 0.7}
NexusMeter.settings.abilityTypeColors[GameLib.CodeEnumDamageType.Heal] = {0, 0.8, 0}
NexusMeter.settings.abilityTypeColors[GameLib.CodeEnumDamageType.Fall] = {0.6, 0.6, 0.6}
NexusMeter.settings.abilityTypeColors[GameLib.CodeEnumDamageType.Suffocate] = {0.6, 0.6, 0.6}

function NexusMeter.GetDefaultWindowSettings()
	local windowDefaultSettings = {}
	windowDefaultSettings.sort = "damage"
	windowDefaultSettings.rows = 8
	windowDefaultSettings.width = 300
	windowDefaultSettings.rowHeight = 18
	windowDefaultSettings.x = (Apollo.GetDisplaySize().nWidth - windowDefaultSettings.width) / 2
	windowDefaultSettings.y = Apollo.GetDisplaySize().nHeight / 2
	return windowDefaultSettings
end

local Ability = {}
Ability.__index = Ability
function Ability:new(info, damageAction)
	local self = {}
	if info.ability:GetId() ~= nil then
	  self.detail = Abilities[info.ability:GetId() .. (info.periodic and "1" or "0")]
	  self.name = self.detail.name
	end
	self.type = info.damagetype
	self.total = 0
	self.totalCrit = 0
	self.max = 0
	self.min = 0
	self.hits = 0
	self.crits = 0
	self.swings = 0
	self.filtered = 0
	self.interrupts = 0

	if damageAction then
		self.deflects = 0
		self.absorbed = 0
		self.intercepted = 0
	end

	if info.owner then
	  self.name = self.name .. " (Pet: " .. info.caster:GetName() .. ")"
	elseif info.periodic then
	  self.name = self.name .. " (Dot)"
	end

	return setmetatable(self, Ability)
end
function Ability:clone()
	local clone = {
		detail = self.detail,
		name = self.name,
		total = self.total,
		type = self.type,
		totalCrit = self.totalCrit,
		max = self.max,
		min = self.min,
		hits = self.hits,
		crits = self.crits,
		swings = self.swings,
		filtered = self.filtered,
		interrupts = self.interrupts
	}

	if self.deflects then -- damageAction
		clone.deflects = self.deflects
		clone.absorbed = self.absorbed
		clone.intercepted = self.intercepted
	end

	return setmetatable(clone, Ability)
end
function Ability:merge(otherAbility)
	self.total = self.total + otherAbility.total
	self.totalCrit = self.totalCrit + otherAbility.totalCrit
	self.max = max(self.max, otherAbility.max)
	if self.min == 0 then
		self.min = otherAbility.min
	else
		if otherAbility.min > 0 then
			self.min = min(self.min, otherAbility.min)
		end
	end
	self.hits = self.hits + otherAbility.hits
	self.crits = self.crits + otherAbility.crits
	self.swings = self.swings + otherAbility.swings
	self.filtered = self.filtered + otherAbility.filtered
	self.interrupts = self.interrupts + otherAbility.interrupts

	if otherAbility.deflects then
		self.deflects = self.deflects + otherAbility.deflects
		self.absorbed = self.absorbed + otherAbility.absorbed
		self.intercepted = self.intercepted + otherAbility.intercepted
	end
end
function Ability:add(statType, amount, info, damageAction)
	self.total = self.total + amount

	if info.crit then
		self.crits = self.crits + 1
		self.totalCrit = self.totalCrit + amount
	elseif statType == "deflects" or statType == "interrupts" then
		self[statType] = self[statType] + 1
	else
		self.hits = self.hits + 1
	end

	self.swings = self.swings + 1

	if amount > self.max then
		self.max = amount
	end

	if amount < self.min or self.min == 0 then
		self.min = amount
	end

	if damageAction then
		self.absorbed = self.absorbed + (info.damageAbsorbed or 0)
		self.intercepted = self.intercepted + (info.damageIntercepted or 0)
	end
end
function Ability:getPreparedAbilityStatData()
	local stats = {
		{ name = L["total"], value = NexusMeter.numberFormat(self.total) },
		{ name = L["max"], value = NexusMeter.numberFormat(self.max) },
		{ name = L["average"], value = NexusMeter.numberFormat(round(self.total / self.swings)) },
		{ name = L["average crit"], value = NexusMeter.numberFormat(round(self.totalCrit / max(self.crits, 1))) },
		{ name = L["min"], value = NexusMeter.numberFormat(self.min) },
		{ name = L["crit rate"], value = ("%.2f%%"):format(self.crits / self.swings * 100) },
		{ name = L["swings"], value = NexusMeter.numberFormat(self.swings) },
		{ name = L["hits"], value = NexusMeter.numberFormat(self.hits) },
		{ name = L["crits"], value = NexusMeter.numberFormat(self.crits) },
    { name = L["filtered"], value = NexusMeter.numberFormat(self.filtered) },
    { name = L["interrupts"], value = NexusMeter.numberFormat(self.interrupts) }
	}

	if self.deflects then
		tinsert(stats, { name = L["deflects"], value = NexusMeter.numberFormat(self.deflects) })
		tinsert(stats, { name = L["absorbed"], value = NexusMeter.numberFormat(self.absorbed) })
		tinsert(stats, { name = L["intercepted"], value = NexusMeter.numberFormat(self.intercepted) })
	end

	return stats
end

local AbilityDetail = {}
function AbilityDetail:new(info)
	local self = {}
	self.name = info:GetName()
	self.icon = ""
	self.type = info.type or "none"
	self.filter = false
	
	if FilteredAbilities[self.name] and not info.abilityNew then
		self.filter = true
	end

	return self
end

local Unit = {}
function Unit:new(detail)
	local self = {}
	self.name = detail:GetName()
	self.player = detail:IsACharacter()
	self.id = detail:GetId()
	--self.inGroup = detail:IsInYourGroup() or detail:IsThePlayer()
	self.self = false
	self.isPet = false
	self.owner = nil

	self.calling = detail:GetClassId()
	self.hostile = (GameLib.GetPlayerUnit():GetDispositionTo(detail) ~= ApolloUnit.CodeEnumDisposition.Friendly)
		
	return self
end

local Player = {}
Player.__index = Player
function Player:new(id, reduced)
	local self = {}
	self.detail = Units[id]
	self.reduced = reduced
	self.damage = 0
	self.damageTaken = 0
	self.damageAbsorbed = 0
	self.friendlyFire = 0
	self.overkill = 0
	self.heal = 0
	self.healTaken = 0
	self.overheal = 0
	self.interrupts = 0

	if reduced then
		self.abilities = {}
		self.linkedToOwner = false
		self.pets = false
		self.interactions = false
	else
		self.linkedToOwner = false
		self.pets = {}

		self.interactions = {
			damage = {},
			damageTaken = {},
			friendlyFire = {},
			overkill = {},
			heal = {},
			healTaken = {},
			overheal = {},
			interrupts = {}
		}
	end

	return setmetatable(self, Player)
end
function Player:addStat(interactedWith, statType, stat, amount, info, damageAction)
	local ability = self:addAbility(interactedWith, statType, stat, amount, info, damageAction)

	if not ability.filter then
		self[statType] = self[statType] + amount
	end

	if statType == "damageTaken" then
		self.damageAbsorbed = self.damageAbsorbed + (info.damageAbsorbed or 0)
	end
end
function Player:addAbility(interactedWith, statType, stat, amount, info, damageAction)
	local abilityDetail = Abilities[info.ability:GetId() .. (info.periodic and "1" or "0")]
	if self.reduced then
		local ability = self.abilities[abilityDetail]
		if not ability then
			ability = Ability:new(info, damageAction)
			self.abilities[abilityDetail] = ability
		end
		ability:add(stat, amount, info, damageAction)
	end

	if not self.reduced and self.interactions then
		local key = interactedWith and interactedWith or L["Unknown"]
		local player = self.interactions[statType][key]
		if not player then
			player = self:new(interactedWith and interactedWith.id, true)
			if not player.detail then
				player.detail = { name = key }
			end
			self.interactions[statType][key] = player
		end

		player:addStat(interactedWith, statType, stat, amount, info, damageAction)
	end

	return abilityDetail
end
function Player:getStat(sort)
	local stat = self[sort]
	for i, value in ipairs(self.pets) do
		stat = stat + value[sort]
	end
	return stat
end
function Player:getInteractions(sort)
	local data = {
		interactions = {},
		count = 0,
		max = 1,
		total = 1
	}

	local interactions = {}
	for _, interaction in pairs(self.interactions[sort]) do
		tinsert(interactions, interaction)
	end

	tsort(interactions, function (a, b)
		return a[sort] > b[sort]
	end)

	for i, interaction in ipairs(interactions) do
		local value = interaction[sort]
		local name = interaction.detail.name

		tinsert(data.interactions, {
			name = name,
			value = value,
			ref = interaction
		})

		data.total = data.total + value
		data.count = data.count + 1
	end

	if data.count > 0 then
		data.max = max(data.interactions[1].value, 1)
	end

	return data
end
function Player:getAbility(ability, sort)
	local returnAbility
	for _, interaction in pairs(self.interactions[sort]) do
		for _, ability2 in pairs(interaction.abilities) do
				if (NexusMeter.settings.mergeAbilitiesByName and (ability.name == ability2.name))
				or (not NexusMeter.settings.mergeAbilitiesByName and (ability.detail == ability2.detail)) then
				if not returnAbility then
					returnAbility = ability2:clone()
				else
					returnAbility:merge(ability2)
				end
			end
		end
	end
	return returnAbility
end
function Player:getAbilities(sort)
	local abilities = {}
	for _, interaction in pairs(self.interactions[sort]) do
		for _, ability in pairs(interaction.abilities) do
			local found = false
			for i, insertedAbility in ipairs(abilities) do
				if (NexusMeter.settings.mergeAbilitiesByName and (insertedAbility.name == ability.name))
					or (not NexusMeter.settings.mergeAbilitiesByName and (insertedAbility.detail == ability.detail)) then
					found = true
					insertedAbility:merge(ability)
				end
			end
			if not found then
				tinsert(abilities, ability:clone())
			end
		end
	end
	
	for _, pet in pairs(self.pets) do
    for _, interaction in pairs(pet.interactions[sort]) do
      for _, ability in pairs(interaction.abilities) do
        local found = false
        for i, insertedAbility in ipairs(abilities) do
          if (NexusMeter.settings.mergeAbilitiesByName and (insertedAbility.name == ability.name))
            or (not NexusMeter.settings.mergeAbilitiesByName and (insertedAbility.detail == ability.detail)) then
            found = true
            insertedAbility:merge(ability)
          end
        end
        if not found then
          tinsert(abilities, ability:clone())
        end
      end
    end
  end

	tsort(abilities, function (a, b)
		return a.total > b.total
	end)

	return abilities
end
function Player:getPreparedAbilityData(sort)
	local data = {
		abilities = {},
		count = 0,
		max = 1,
		total = 0
	}

	local abilities = self:getAbilities(sort)

	for i, ability in ipairs(abilities) do
		local value = ability.total

		tinsert(data.abilities, {
			value = value,
			ref = ability
		})

		data.total = data.total + value
		data.count = data.count + 1
	end

	data.total = max(data.total, 1)
	if data.count > 0 then
		data.max = max(data.abilities[1].value, 1)
	end

	return data
end
function Player:getInteractionAbilityData()
	local data = {
		abilities = {},
		count = 0,
		max = 1,
		total = 0
	}

	local abilities = {}
	for _, ability in pairs(self.abilities) do
		tinsert(abilities, ability)
	end

	tsort(abilities, function (a, b)
		return a.total > b.total
	end)

	for i, ability in ipairs(abilities) do
		local value = ability.total

		tinsert(data.abilities, {
			value = value,
			ref = ability
		})

		data.total = data.total + value
		data.count = data.count + 1
	end

	data.total = max(data.total, 1)
	if data.count > 0 then
		data.max = max(data.abilities[1].value, 1)
	end

	return data
end
function Player:createFakeAbility()
	local fakeAbility = Ability:new({ability = {GetId = function() return nil end}}, true)
	local backup = fakeAbility.getPreparedAbilityStatData
	local player = self
	fakeAbility.detail = AbilityDetail:new({GetName = function() return L["Total"] end})
	fakeAbility.name = L["Total"]

	function fakeAbility:getPreparedAbilityStatData(sort)
		-- force update on retrieve

		self.total = 0
		self.totalCrit = 0
		self.max = 0
		self.min = 0
		self.hits = 0
		self.crits = 0
		self.swings = 0
		self.filtered = 0
		self.interrupts = 0

		self.deflects = 0
		self.absorbed = 0
		self.intercepted = 0

		if player.interactions then -- real player
			for _, interaction in pairs(player.interactions[sort]) do
				for _, ability in pairs(interaction.abilities) do
					self:merge(ability)
				end
			end
		else -- interacted player
			for _, ability in pairs(player.abilities) do
				self:merge(ability)
			end
		end

		return backup(self)
	end

	return fakeAbility
end
function Player:getTooltip(sort)
	local result = {"<P TextColor=\"FFFFD100\">" .. L["Top 3 Abilities:"] .. "</P>"}

	local players = {}
	for _, player in pairs(self.interactions[sort]) do
		tinsert(players, player)
	end

	tsort(players, function(a, b)
		return a[sort] > b[sort]
	end)

	local abilities = self:getAbilities(sort)

	for i = 1, 3 do
		if abilities[i] then
			tinsert(result, ("   (%d%%) %s"):format(abilities[i].total / self[sort] * 100, abilities[i].name:sub(0, 16)))
		end
	end

	tinsert(result, "<P TextColor=\"FFFFD100\">" .. L["Top 3 Interactions:"] .. "</P>")

	for i = 1, 3 do
		if players[i] then
			tinsert(result, ("   (%d%%) %s"):format(players[i][sort] / self[sort] * 100, players[i].detail.name:sub(0, 16)))
		end
	end

	tinsert(result, "<P TextColor=\"FF33FF33\">&lt;" .. L["Middle-Click for interactions"] .. "&gt;</P>")

	return table.concat(result, "\n")
end


local Combat = {}
Combat.__index = Combat
function Combat:new(overall)
	local self = {}
	self.startTime = GameLib.GetGameTime()
	self.overall = overall
	self.duration = 0
	self.previousDuration = 0
	self.players = {}
	self.hostiles = {}
	self.hasBoss = false
	return setmetatable(self, Combat)
end
function Combat:endCombat(durationIsCallTime)
	local now = GameLib.GetGameTime()
	if durationIsCallTime then
		self.duration = now - self.startTime
	else
		self.duration = now - self.startTime - (now - LastDamageAction)
	end
end
function Combat:getPreparedPlayerData(sort, showNpcs)
	local data = {
		players = {},
		count = 0,
		max = 0,
		total = 0
	}

	local players = {}
	for _, player in pairs(self.players) do
		if (showNpcs and not player.detail.player and not player.detail.isPet)
			or (not showNpcs and (player.detail.player and not player.detail.isPet)) then
			tinsert(players, player)
		end
	end

	-- Players and pets > other units + sort by stat desc
	tsort(players, function (a, b)
		return a:getStat(sort) > b:getStat(sort)
	end)

	for i, player in ipairs(players) do
		local value = player[sort]
		local valuePlusPets = player:getStat(sort)
		if valuePlusPets > 0 then
			local name
			if player.linkedToOwner then
				name = ("%s (%s)"):format(player.detail.name, player.detail.owner.name)
			else
				name = player.detail.name
			end

			tinsert(data.players, {
				name = name,
				value = valuePlusPets,
				ref = player
			})

			data.total = data.total + value
			data.count = data.count + 1
		end
	end

	data.total = data.total
	if data.count > 0 then
		data.max = data.players[1].value
	end

	return data
end
function Combat:getHostile()
	local players = {}
	
	if self.overall then
	   return L["Total"]
	end
	
	for _, player in pairs(self.players) do
		tinsert(players, player)
	end

	-- other units > Players and pets + sort by damage taken desc
	tsort(players, function (a, b)
		local aCond = not (a.detail.player or a.detail.isPet)
		local bCond = not (b.detail.player or b.detail.isPet)
		if aCond and bCond then
			return a.damageTaken > b.damageTaken
		end
		return aCond and not bCond
	end)

	if #players > 0 then
		return players[1].detail.name
	end

	return ""
end
function Combat:addPlayer(id)
	local unit = Units[id]
	local player = self.players[unit]
	if not player then
		player = Player:new(id, false)
		self.players[unit] = player
	end
	if player.detail.isPet then
		self:linkPetToOwner(player)
	end

	return player
end
function Combat:linkPetToOwner(pet)
	if not pet.linkedToOwner then
		local owner = self.players[pet.detail.owner]
		if owner then
			pet.linkedToOwner = true
			tinsert(owner.pets, pet)
		end
	end
end


local function AddGlobalAbility(info)
	local key = info.ability:GetId() .. (info.periodic and "1" or "0") 
	local ability = Abilities[key]
	if not ability then
		Abilities[key] = {
		  name = info.ability:GetName(),
		  icon = info.ability:GetIcon(),
		}
	end
	return ability
end

local function AddGlobalUnit(detail, owner)
	local id = detail:GetId()
	local unit = Units[id]
	if not unit then
		unit = Unit:new(detail)
		
		-- search for pet owner if pet
		if owner then -- friendly?
			unit.isPet = true
			unit.owner = AddGlobalUnit(owner, nil)
			--unit.inGroup = unit.owner.inGroup
		end
		
		if detail:IsThePlayer() then
			unit.self = true
		end
		
		Units[id] = unit
		return unit
	end

	return unit
end

local function GetMaxValueCombat(sort, showNpcs)
	local maxvalue = 1
	for i, combat in ipairs(NexusMeter.combats) do
		local value = combat:getPreparedPlayerData(sort, showNpcs).total / max(combat.duration, 1)
		if value > maxvalue then
			maxvalue = value
		end
	end
	return maxvalue
end

local function NewCombat(permanent)
	if InCombat then
		return
	end

  if not NexusMeter.overallCombat then
    NexusMeter.overallCombat = Combat:new(true)
    tinsert(NexusMeter.combats, NexusMeter.overallCombat)
  end

	NexusMeter.CurrentCombat = Combat:new(false)
	InCombat = true
	Permanent = not not permanent
	tinsert(NexusMeter.combats, NexusMeter.CurrentCombat)

	NexusMeter.UI.NewCombat()
end

local function EndCombat(durationIsCallTime)
  if not InCombat then
    return
	end

  NexusMeter.CurrentCombat:endCombat(durationIsCallTime)
  NexusMeter.overallCombat.previousDuration = NexusMeter.overallCombat.previousDuration + NexusMeter.CurrentCombat.duration
  NexusMeter.overallCombat.duration = NexusMeter.overallCombat.previousDuration
	StopTracking = false
	InCombat = false
	Permanent = false

	NexusMeter.UI.Update()
	NexusMeter.UI.EndCombat()
end

local function CombatEventsHandler(info, statType, damageAction)
	if StopTracking and not Permanent then -- force a combat end on several bosses with incoming damage after kill extending the duration
		return
	end

  -- Sometimes the API gives us a nil target or caster, ignore these cases.
	if not info.caster or not info.target then return end

	local target = AddGlobalUnit(info.target, nil)
	local caster = AddGlobalUnit(info.caster, info.owner)

	AddGlobalAbility(info)

	local selfAction = info.caster == info.target
	if not selfAction then
		if statType == "damage" or statType == "interrupts" then
			if not InCombat then --and caster.inGroup then
				NewCombat()
			end
		end
	else
		-- self damage action, e.g. fall damage
		if statType == "damage" then
			statType = "damageTaken"
		end
	end

	if InCombat then
		if damageAction then-- and caster.inGroup then
			LastDamageAction = GameLib.GetGameTime()
			NeedsUpdate = true
		end
	else
		return
	end
	
	if info.caster:GetRank() == ApolloUnit.CodeEnumRank.Elite or 
     info.target:GetRank() == ApolloUnit.CodeEnumRank.Elite then
    NexusMeter.CurrentCombat.hasBoss = true
  end

	local amount = info[statType] or 0


  local overallCasterInCombat = NexusMeter.overallCombat:addPlayer(info.caster:GetId())
	local casterInCombat = NexusMeter.CurrentCombat:addPlayer(info.caster:GetId())

	-- Add stat to caster
	local stat
	if statType == "damageTaken" then
		stat = "damageTaken"
		amount = info.damage or 0
	elseif damageAction then
		stat = "damage"
	else
		stat = "heal"
	end
	
	overallCasterInCombat:addStat(target, stat, statType, amount, info, damageAction)
	casterInCombat:addStat(target, stat, statType, amount, info, damageAction)

	-- Add Overheal/Overkill
	local key
	if info.overkill and info.overkill > 0 then
		key = "overkill"
	elseif info.overheal and info.overheal > 0 then
		key = "overheal"
	elseif info.interrupts and info.interrupts > 0 then
	  key = "interrupts"
	end

	if key then
	  overallCasterInCombat:addStat(target, key, key, info[key], info, damageAction)
		casterInCombat:addStat(target, key, key, info[key], info, damageAction)
	end

	-- Add targets equivalent damage/heal taken
	if target then
	  local overallTargetInCombat = NexusMeter.overallCombat:addPlayer(info.target:GetId())
		local targetInCombat = NexusMeter.CurrentCombat:addPlayer(info.target:GetId())

		local taken
		if damageAction then
			taken = "damageTaken"
		elseif statType == "heal" then
			taken = "healTaken"
		end

		if taken then
		  overallTargetInCombat:addStat(caster, taken, taken, amount, info, damageAction)
			targetInCombat:addStat(caster, taken, taken, amount, info, damageAction)
		end
	end
end

-- TODO: This is for tracking mob deaths for EndCombatAfterKill and NewCombatAfterKill, don't need yet
-- TODO: Add death log
--function Events.Death(info)
--	if not InCombat or StopTracking or not info.target then
--		return
--	end
--
--	local detail = InspectUnitDetail(info.target)
--	if not detail or not detail.type then
--		return
--	end
--	for i, type in ipairs(EndCombatAfterKill) do
--		if type == detail.type then
--			EndCombat()
--			StopTracking = true
--			InCombat = true
--			return
--		end
--	end
--	for i, type in ipairs(NewCombatAfterKill) do
--		if type == detail.type then
--			EndCombat()
--			return
--		end
--	end
--end

local function On()
	if NexusMeter.settings.enabled then
		return
	end

	NexusMeter.settings.enabled = true
  NexusMeter.UI.Visible(true)

	Apollo.RegisterEventHandler("CombatLogHeal",					  "OnCombatLogHeal", NexusMeter)
	Apollo.RegisterEventHandler("CombatLogDamage",					"OnCombatLogDamage", NexusMeter)
	Apollo.RegisterEventHandler("CombatLogDeflect",					"OnCombatLogDeflect", NexusMeter)
  Apollo.RegisterEventHandler("CombatLogTransference",    "OnCombatLogTransference", NexusMeter)
  Apollo.RegisterEventHandler("CombatLogCCState",         "OnCombatLogCCState", NexusMeter)
  
	NexusMeter.timerPulse:Start()

	--RegisterEvent(Event.Combat.Death, 				{Events.Death,  		Info.identifier, "Handle Deaths"})
end

local function Off()
	if not NexusMeter.settings.enabled then
		return
	end
	
	NexusMeter.settings.enabled = false
	NexusMeter.UI.Visible(false)

	Apollo.RemoveEventHandler("CombatLogHeal")
	Apollo.RemoveEventHandler("CombatLogDamage")
	Apollo.RemoveEventHandler("CombatLogDeflect")
	Apollo.RegisterEventHandler("CombatLogTransference")
  Apollo.RegisterEventHandler("CombatLogCCState")
	NexusMeter.timerPulse:Stop()

	--UnregisterEvent(Event.Combat.Death, 		Events.Death)
end

local function Toggle()
	if NexusMeter.settings.enabled then
		Off()
	else
		On()
	end
end

local function Reset()
  EndCombat()

  Units = {}
  Abilities = {}
	NexusMeter.CurrentCombat = {}
	NexusMeter.overallCombat = nil

	InCombat = false
	NeedsUpdate = false

	NexusMeter.combats = {}

	NexusMeter.UI.Reset()
end

local SlashCommands = setmetatable({
	show = function ()
		On()
	end,
	hide = function ()
		Off()
	end,
	default = function ()
		NexusMeter.UI.Default()
	end,
	lock = function ()
	  NexusMeter.settings.lock = true
	  NexusMeter.UI.ShowResizer(false)
	end,
	unlock = function ()
    NexusMeter.settings.lock = false
    NexusMeter.UI.ShowResizer(true)
  end,
	config = function ()
		NexusMeter.ConfigInit()
	end,
	toggle = function ()
		if NexusMeter.settings.enabled then
			Off()
		else
			On()
		end
	end
},
{
	__index = function (t, key)
		return function ()
			Print(L["Available commands:"])
			for cmd, func in pairs(t) do
				Print("- " .. cmd)
			end
		end
	end
})

function NexusMeter:OnLoad()
  -- TODO: yuk..
	L = NexusMeter.l

  -- Set console variables to allow logging to work properly
  Apollo.SetConsoleVariable("cmbtlog.disableOtherPlayers", false)

  -- Load Sprites
  Apollo.LoadSprites("NexusMeterSprites.xml", "NexusMeterSprites")
  
  -- Load Forms
  self.xmlMainDoc = XmlDoc.CreateFromFile("NexusMeter.xml")
  self.xmlMainDoc:RegisterCallback("OnDocLoaded", self)
end

function NexusMeter:OnDocLoaded()
  Apollo.RegisterEventHandler("WindowManagementReady",    "OnWindowManagementReady", self)
end  
  
function NexusMeter:OnSave(eType)
  if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then return end
  
  return NexusMeter.deepcopy(self.settings)
end

function NexusMeter:OnRestore(eType, tSavedState)
  if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then return end
  
  self.default_settings = NexusMeter.deepcopy(self.settings)
  
  self.settings = NexusMeter.deepcopy(tSavedState)
  
  -- TODO: This needs to be a deep check
  for key, value in pairs(self.default_settings) do
    if self.settings[key] == nil then
      self.settings[key] = value
    end
  end
  
  -- Forcibly override class colors since we got them wrong. (This isn't configurable, so it doesn't matter for now).
  self.settings.classColors = self.default_settings.classColors
  self.settings.abilityTypeColors = self.default_settings.abilityTypeColors
end

function NexusMeter:OnWindowManagementReady()
  -- Create update timer
  self.timerPulse = ApolloTimer.Create(0.1, true, "Update", self)
  
  -- Register slash handler
  Apollo.RegisterSlashCommand("nm", "SlashHandler", self)

  -- If we have no window settings, then add default window
  if #NexusMeter.settings.windows == 0 then
    tinsert(NexusMeter.settings.windows, NexusMeter.GetDefaultWindowSettings())
  end
  
  -- TODO: yuk..
  NexusMeter.EndCombat = EndCombat  
  NexusMeter.Reset = Reset
  NexusMeter.On = On
  NexusMeter.Off = Off
  NexusMeter.NewCombat = NewCombat
  NexusMeter.GetMaxValueCombat = GetMaxValueCombat
  
  -- Create UI windows
  NexusMeter.UI.Init()
  
  local loadui = NexusMeter.settings.enabled
  NexusMeter.settings.enabled = false
  
  -- Turn events on
  if loadui then
    On()
  else
    Print((L["Type /nm show to reactivate %s."]):format(Apollo.GetAddonInfo("NexusMeter").strName))
    Off()
  end
end

function NexusMeter:OnCombatLogHeal(tEventArgs)
	local info = {
		target = tEventArgs.unitTarget,
		caster = tEventArgs.unitCaster,
		owner = tEventArgs.unitCasterOwner,
		ability = tEventArgs.splCallingSpell,
		damagetype = GameLib.CodeEnumDamageType.Heal,
		heal = tEventArgs.nHealAmount - tEventArgs.nOverheal,
		overheal = tEventArgs.nOverheal,
		crit = tEventArgs.eCombatResult == GameLib.CodeEnumCombatResult.Critical,
	}
	
	CombatEventsHandler(info, "heal", false)
end

function NexusMeter:OnCombatLogCCState(tEventArgs)
  if tEventArgs.nInterruptArmorHit > 0 then
    local info = {
      target = tEventArgs.unitTarget,
      caster = tEventArgs.unitCaster,
      owner = tEventArgs.unitCasterOwner,
      ability = tEventArgs.splCallingSpell,
      interrupts = tEventArgs.nInterruptArmorHit,
    }
    
    CombatEventsHandler(info, "interrupts", true)
  end
end

function NexusMeter:OnCombatLogDeflect(tEventArgs)
	local info = {
		target = tEventArgs.unitTarget,
		caster = tEventArgs.unitCaster,
		owner = tEventArgs.unitCasterOwner,
		ability = tEventArgs.splCallingSpell,
	}

	CombatEventsHandler(info, "deflects", true)
end

function NexusMeter:OnCombatLogTransference(tEventArgs)
  local info = {
    target = tEventArgs.unitTarget,
    caster = tEventArgs.unitCaster,
    owner = tEventArgs.unitCasterOwner,
    ability = tEventArgs.splCallingSpell,
    damagetype = tEventArgs.eDamageType,
    damage = tEventArgs.nDamageAmount - tEventArgs.nOverkill,
    periodic = tEventArgs.bPeriodic,
    overkill = tEventArgs.nOverkill,
    crit = tEventArgs.eCombatResult == GameLib.CodeEnumCombatResult.Critical,
  }
  
  -- Work around an issue where overkill == damage, in this case there is actually a bug in the data given to us
  -- and we didn't have purely overkill. In this case don't subtract the overkill damage.
  if tEventArgs.nDamageAmount == tEventArgs.nOverkill then
    info.damage = tEventArgs.nDamageAmount
  end 

  CombatEventsHandler(info, "damage", true)
  
  local info = {
    target = tEventArgs.unitTarget,
    caster = tEventArgs.unitCaster,
    owner = tEventArgs.unitCasterOwner,
    ability = tEventArgs.splCallingSpell,
    damagetype = GameLib.CodeEnumDamageType.Heal,
    heal = tEventArgs.tHealData[1].nHealAmount - tEventArgs.tHealData[1].nOverheal,
    overheal = tEventArgs.tHealData[1].nOverheal,
    crit = tEventArgs.eCombatResult == GameLib.CodeEnumCombatResult.Critical,
  }
  
  CombatEventsHandler(info, "heal", false)
end

function NexusMeter:OnCombatLogDamage(tEventArgs)
	local info = {
		target = tEventArgs.unitTarget,
		caster = tEventArgs.unitCaster,
		owner = tEventArgs.unitCasterOwner,
		ability = tEventArgs.splCallingSpell,
		damagetype = tEventArgs.eDamageType,
		damage = tEventArgs.nDamageAmount - tEventArgs.nOverkill,
		periodic = tEventArgs.bPeriodic,
		overkill = tEventArgs.nOverkill,
		crit = tEventArgs.eCombatResult == GameLib.CodeEnumCombatResult.Critical,
	}
	
	if info.target and
	  (info.target:GetName() == "Target Dummy" or
	   info.target:GetName() == "Formidable Target Dummy" or
     info.target:GetName() == "Weak Target Dummy") then
		info.damage = tEventArgs.nRawDamage - tEventArgs.nOverkill
	end
	
	-- Work around an issue where overkill == damage, in this case there is actually a bug in the data given to us
	-- and we didn't have purely overkill. In this case don't subtract the overkill damage.
	if tEventArgs.nDamageAmount == tEventArgs.nOverkill then
    info.damage = tEventArgs.nDamageAmount
  end 

	CombatEventsHandler(info, "damage", true)
end

function NexusMeter:Update()
  local now = GameLib.GetGameTime()

  if InCombat then
    if now - LastDamageAction >= 6 and not Permanent then
      EndCombat()
		end

		if now - LastTimerUpdate > 1 then
			LastTimerUpdate = now
			NexusMeter.UI.TimerUpdate(now - NexusMeter.CurrentCombat.startTime)
		end
	end

	if not NeedsUpdate and not Permanent then return end

	if now - LastUpdate > 0.3 then
	  NexusMeter.overallCombat.duration = NexusMeter.overallCombat.previousDuration + now - NexusMeter.CurrentCombat.startTime
		NexusMeter.CurrentCombat.duration = now - NexusMeter.CurrentCombat.startTime
		LastUpdate = now
		NeedsUpdate = false

		NexusMeter.UI.Update()
	end
end

function NexusMeter:SlashHandler(cmd, arg)
	local list = {}

	for param in arg:gmatch("[^%s]+") do
		tinsert(list, param)
	end
	SlashCommands[list[1]](unpack(list, 2, #list))
end

-- Register addon
Apollo.RegisterAddon(NexusMeter)